/* Copyright (C) 2010-2011 Linaro Ltd.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

/*
 * Authors:
 *  Jammy Zhou <jammy.zhou@linaro.org>
 *  Marc Ordinas i Llopis, Collabora Ltd <marc.ordinasillopis@linaro.org>
 */

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <dlfcn.h>
#include <string.h>
#include <unistd.h>
#include "GL/gl_proxy.h"
#include "GL/glx_proxy.h"
#include "EGL/egl_proxy.h"

/* Struct where glproxy keeps things. */
struct glProxyContext
{
    GLuint current_backends;
    void *gles2_handle;
    void *gl_handle;
    void *egl_handle;

    GLboolean done_detect_libraries;
    GLuint detected_libraries;
    GLuint num_detected_backends;
    glProxyBackendInfo detected_backends[3];
};

/* library names for different backends */
#define LIBGL_NAME      "libGL.so.1"
#define LIBGLES2_NAME   "libGLESv2.so.2"
#define LIBEGL_NAME     "libEGL.so.1"

/* proxy functions */
PFNPROXYACTIVETEXTUREPROC pfnProxyActiveTexture = NULL;
PFNPROXYATTACHSHADERPROC pfnProxyAttachShader = NULL;
PFNPROXYBINDATTRIBLOCATIONPROC pfnProxyBindAttribLocation = NULL;
PFNPROXYBINDBUFFERPROC pfnProxyBindBuffer = NULL;
PFNPROXYBINDFRAMEBUFFERPROC pfnProxyBindFramebuffer = NULL;
PFNPROXYBINDRENDERBUFFERPROC pfnProxyBindRenderbuffer = NULL;
PFNPROXYBINDTEXTUREPROC pfnProxyBindTexture = NULL;
PFNPROXYBLENDCOLORPROC pfnProxyBlendColor = NULL;
PFNPROXYBLENDEQUATIONPROC pfnProxyBlendEquation = NULL;
PFNPROXYBLENDEQUATIONSEPARATEPROC pfnProxyBlendEquationSeparate = NULL;
PFNPROXYBLENDFUNCPROC pfnProxyBlendFunc = NULL;
PFNPROXYBLENDFUNCSEPARATEPROC pfnProxyBlendFuncSeparate = NULL;
PFNPROXYBUFFERDATAPROC pfnProxyBufferData = NULL;
PFNPROXYBUFFERSUBDATAPROC pfnProxyBufferSubData = NULL;
PFNPROXYCHECKFRAMEBUFFERSTATUSPROC pfnProxyCheckFramebufferStatus = NULL;
PFNPROXYCLEARPROC pfnProxyClear = NULL;
PFNPROXYCLEARCOLORPROC pfnProxyClearColor = NULL;
PFNPROXYCLEARDEPTHFPROC pfnProxyClearDepthf = NULL;
PFNPROXYCLEARSTENCILPROC pfnProxyClearStencil = NULL;
PFNPROXYCOLORMASKPROC pfnProxyColorMask = NULL;
PFNPROXYCOMPILESHADERPROC pfnProxyCompileShader = NULL;
PFNPROXYCOMPRESSEDTEXIMAGE2DPROC pfnProxyCompressedTexImage2D = NULL;
PFNPROXYCOMPRESSEDTEXSUBIMAGE2DPROC pfnProxyCompressedTexSubImage2D = NULL;
PFNPROXYCOPYTEXIMAGE2DPROC pfnProxyCopyTexImage2D = NULL;
PFNPROXYCOPYTEXSUBIMAGE2DPROC pfnProxyCopyTexSubImage2D = NULL;
PFNPROXYCREATEPROGRAMPROC pfnProxyCreateProgram = NULL;
PFNPROXYCREATESHADERPROC pfnProxyCreateShader = NULL;
PFNPROXYCULLFACEPROC pfnProxyCullFace = NULL;
PFNPROXYDELETEBUFFERSPROC pfnProxyDeleteBuffers = NULL;
PFNPROXYDELETEFRAMEBUFFERSPROC pfnProxyDeleteFramebuffers = NULL;
PFNPROXYDELETEPROGRAMPROC pfnProxyDeleteProgram = NULL;
PFNPROXYDELETERENDERBUFFERSPROC pfnProxyDeleteRenderbuffers = NULL;
PFNPROXYDELETESHADERPROC pfnProxyDeleteShader = NULL;
PFNPROXYDELETETEXTURESPROC pfnProxyDeleteTextures = NULL;
PFNPROXYDEPTHFUNCPROC pfnProxyDepthFunc = NULL;
PFNPROXYDEPTHMASKPROC pfnProxyDepthMask = NULL;
PFNPROXYDEPTHRANGEFPROC pfnProxyDepthRangef = NULL;
PFNPROXYDETACHSHADERPROC pfnProxyDetachShader = NULL;
PFNPROXYDISABLEPROC pfnProxyDisable = NULL;
PFNPROXYDISABLEVERTEXATTRIBARRAYPROC pfnProxyDisableVertexAttribArray = NULL;
PFNPROXYDRAWARRAYSPROC pfnProxyDrawArrays = NULL;
PFNPROXYDRAWELEMENTSPROC pfnProxyDrawElements = NULL;
PFNPROXYENABLEPROC pfnProxyEnable = NULL;
PFNPROXYENABLEVERTEXATTRIBARRAYPROC pfnProxyEnableVertexAttribArray = NULL;
PFNPROXYFINISHPROC pfnProxyFinish = NULL;
PFNPROXYFLUSHPROC pfnProxyFlush = NULL;
PFNPROXYFRAMEBUFFERRENDERBUFFERPROC pfnProxyFramebufferRenderbuffer = NULL;
PFNPROXYFRAMEBUFFERTEXTURE2DPROC pfnProxyFramebufferTexture2D = NULL;
PFNPROXYFRONTFACEPROC pfnProxyFrontFace = NULL;
PFNPROXYGENBUFFERSPROC pfnProxyGenBuffers = NULL;
PFNPROXYGENERATEMIPMAPPROC pfnProxyGenerateMipmap = NULL;
PFNPROXYGENFRAMEBUFFERSPROC pfnProxyGenFramebuffers = NULL;
PFNPROXYGENRENDERBUFFERSPROC pfnProxyGenRenderbuffers = NULL;
PFNPROXYGENTEXTURESPROC pfnProxyGenTextures = NULL;
PFNPROXYGETACTIVEATTRIBPROC pfnProxyGetActiveAttrib = NULL;
PFNPROXYGETACTIVEUNIFORMPROC pfnProxyGetActiveUniform = NULL;
PFNPROXYGETATTACHEDSHADERSPROC pfnProxyGetAttachedShaders = NULL;
PFNPROXYGETATTRIBLOCATIONPROC pfnProxyGetAttribLocation = NULL;
PFNPROXYGETBOOLEANVPROC pfnProxyGetBooleanv = NULL;
PFNPROXYGETBUFFERPARAMETERIVPROC pfnProxyGetBufferParameteriv = NULL;
PFNPROXYGETERRORPROC pfnProxyGetError = NULL;
PFNPROXYGETFLOATVPROC pfnProxyGetFloatv = NULL;
PFNPROXYGETFRAMEBUFFERATTACHMENTPARAMETERIVPROC pfnProxyGetFramebufferAttachmentParameteriv = NULL;
PFNPROXYGETINTEGERVPROC pfnProxyGetIntegerv = NULL;
PFNPROXYGETPROGRAMIVPROC pfnProxyGetProgramiv = NULL;
PFNPROXYGETPROGRAMINFOLOGPROC pfnProxyGetProgramInfoLog = NULL;
PFNPROXYGETRENDERBUFFERPARAMETERIVPROC pfnProxyGetRenderbufferParameteriv = NULL;
PFNPROXYGETSHADERIVPROC pfnProxyGetShaderiv = NULL;
PFNPROXYGETSHADERINFOLOGPROC pfnProxyGetShaderInfoLog = NULL;
PFNPROXYGETSHADERPRECISIONFORMATPROC pfnProxyGetShaderPrecisionFormat = NULL;
PFNPROXYGETSHADERSOURCEPROC pfnProxyGetShaderSource = NULL;
PFNPROXYGETSTRINGPROC pfnProxyGetString = NULL;
PFNPROXYGETTEXPARAMETERFVPROC pfnProxyGetTexParameterfv = NULL;
PFNPROXYGETTEXPARAMETERIVPROC pfnProxyGetTexParameteriv = NULL;
PFNPROXYGETUNIFORMFVPROC pfnProxyGetUniformfv = NULL;
PFNPROXYGETUNIFORMIVPROC pfnProxyGetUniformiv = NULL;
PFNPROXYGETUNIFORMLOCATIONPROC pfnProxyGetUniformLocation = NULL;
PFNPROXYGETVERTEXATTRIBFVPROC pfnProxyGetVertexAttribfv = NULL;
PFNPROXYGETVERTEXATTRIBIVPROC pfnProxyGetVertexAttribiv = NULL;
PFNPROXYGETVERTEXATTRIBPOINTERVPROC pfnProxyGetVertexAttribPointerv = NULL;
PFNPROXYHINTPROC pfnProxyHint = NULL;
PFNPROXYISBUFFERPROC pfnProxyIsBuffer = NULL;
PFNPROXYISENABLEDPROC pfnProxyIsEnabled = NULL;
PFNPROXYISFRAMEBUFFERPROC pfnProxyIsFramebuffer = NULL;
PFNPROXYISPROGRAMPROC pfnProxyIsProgram = NULL;
PFNPROXYISRENDERBUFFERPROC pfnProxyIsRenderbuffer = NULL;
PFNPROXYISSHADERPROC pfnProxyIsShader = NULL;
PFNPROXYISTEXTUREPROC pfnProxyIsTexture = NULL;
PFNPROXYLINEWIDTHPROC pfnProxyLineWidth = NULL;
PFNPROXYLINKPROGRAMPROC pfnProxyLinkProgram = NULL;
PFNPROXYPIXELSTOREIPROC pfnProxyPixelStorei = NULL;
PFNPROXYPOLYGONOFFSETPROC pfnProxyPolygonOffset = NULL;
PFNPROXYREADPIXELSPROC pfnProxyReadPixels = NULL;
PFNPROXYRELEASESHADERCOMPILERPROC pfnProxyReleaseShaderCompiler = NULL;
PFNPROXYRENDERBUFFERSTORAGEPROC pfnProxyRenderbufferStorage = NULL;
PFNPROXYSAMPLECOVERAGEPROC pfnProxySampleCoverage = NULL;
PFNPROXYSCISSORPROC pfnProxyScissor = NULL;
PFNPROXYSHADERBINARYPROC pfnProxyShaderBinary = NULL;
PFNPROXYSHADERSOURCEPROC pfnProxyShaderSource = NULL;
PFNPROXYSTENCILFUNCPROC pfnProxyStencilFunc = NULL;
PFNPROXYSTENCILFUNCSEPARATEPROC pfnProxyStencilFuncSeparate = NULL;
PFNPROXYSTENCILMASKPROC pfnProxyStencilMask = NULL;
PFNPROXYSTENCILMASKSEPARATEPROC pfnProxyStencilMaskSeparate = NULL;
PFNPROXYSTENCILOPPROC pfnProxyStencilOp = NULL;
PFNPROXYSTENCILOPSEPARATEPROC pfnProxyStencilOpSeparate = NULL;
PFNPROXYTEXIMAGE2DPROC pfnProxyTexImage2D = NULL;
PFNPROXYTEXPARAMETERFPROC pfnProxyTexParameterf = NULL;
PFNPROXYTEXPARAMETERFVPROC pfnProxyTexParameterfv = NULL;
PFNPROXYTEXPARAMETERIPROC pfnProxyTexParameteri = NULL;
PFNPROXYTEXPARAMETERIVPROC pfnProxyTexParameteriv = NULL;
PFNPROXYTEXSUBIMAGE2DPROC pfnProxyTexSubImage2D = NULL;
PFNPROXYUNIFORM1FPROC pfnProxyUniform1f = NULL;
PFNPROXYUNIFORM1FVPROC pfnProxyUniform1fv = NULL;
PFNPROXYUNIFORM1IPROC pfnProxyUniform1i = NULL;
PFNPROXYUNIFORM1IVPROC pfnProxyUniform1iv = NULL;
PFNPROXYUNIFORM2FPROC pfnProxyUniform2f = NULL;
PFNPROXYUNIFORM2FVPROC pfnProxyUniform2fv = NULL;
PFNPROXYUNIFORM2IPROC pfnProxyUniform2i = NULL;
PFNPROXYUNIFORM2IVPROC pfnProxyUniform2iv = NULL;
PFNPROXYUNIFORM3FPROC pfnProxyUniform3f = NULL;
PFNPROXYUNIFORM3FVPROC pfnProxyUniform3fv = NULL;
PFNPROXYUNIFORM3IPROC pfnProxyUniform3i = NULL;
PFNPROXYUNIFORM3IVPROC pfnProxyUniform3iv = NULL;
PFNPROXYUNIFORM4FPROC pfnProxyUniform4f = NULL;
PFNPROXYUNIFORM4FVPROC pfnProxyUniform4fv = NULL;
PFNPROXYUNIFORM4IPROC pfnProxyUniform4i = NULL;
PFNPROXYUNIFORM4IVPROC pfnProxyUniform4iv = NULL;
PFNPROXYUNIFORMMATRIX2FVPROC pfnProxyUniformMatrix2fv = NULL;
PFNPROXYUNIFORMMATRIX3FVPROC pfnProxyUniformMatrix3fv = NULL;
PFNPROXYUNIFORMMATRIX4FVPROC pfnProxyUniformMatrix4fv = NULL;
PFNPROXYUSEPROGRAMPROC pfnProxyUseProgram = NULL;
PFNPROXYVALIDATEPROGRAMPROC pfnProxyValidateProgram = NULL;
PFNPROXYVERTEXATTRIB1FPROC pfnProxyVertexAttrib1f = NULL;
PFNPROXYVERTEXATTRIB1FVPROC pfnProxyVertexAttrib1fv = NULL;
PFNPROXYVERTEXATTRIB2FPROC pfnProxyVertexAttrib2f = NULL;
PFNPROXYVERTEXATTRIB2FVPROC pfnProxyVertexAttrib2fv = NULL;
PFNPROXYVERTEXATTRIB3FPROC pfnProxyVertexAttrib3f = NULL;
PFNPROXYVERTEXATTRIB3FVPROC pfnProxyVertexAttrib3fv = NULL;
PFNPROXYVERTEXATTRIB4FPROC pfnProxyVertexAttrib4f = NULL;
PFNPROXYVERTEXATTRIB4FVPROC pfnProxyVertexAttrib4fv = NULL;
PFNPROXYVERTEXATTRIBPOINTERPROC pfnProxyVertexAttribPointer = NULL;
PFNPROXYVIEWPORTPROC pfnProxyViewport = NULL;

static void *glProxyOpenLibrary(const char *name)
{
    return dlopen(name, RTLD_LAZY | RTLD_GLOBAL);
}

static int glProxyCloseLibrary(void *handle)
{
    return dlclose(handle);
}

void *glProxyGetProcAddr(void *handle, const char *name) {
    if (pfnProxyGLXGetProcAddress) {
        return (*pfnProxyGLXGetProcAddress)((const GLubyte *) name);
    } else if (pfnProxyGLXGetProcAddressARB) {
        return (*pfnProxyGLXGetProcAddressARB)((const GLubyte *) name);
    } else {
        return dlsym(handle, name);
    }
}

GLboolean glProxyCheckExtension(const char *list, const char *name)
{
    size_t len = strlen(name);
    const char *ext_ptr = list;

    if(ext_ptr == NULL)
        return GL_FALSE;

    while ((ext_ptr = strstr(ext_ptr, name)) != NULL) {
        if (ext_ptr[len] == ' ' || ext_ptr[len] == '\0')
            break;
        ext_ptr += len;
    }

    return (ext_ptr != NULL)? GL_TRUE : GL_FALSE;
}

GLboolean glProxyHasExtension(void *display, const char *extension)
{
    char const *backend_extensions = NULL;
    char const *gl_extensions = NULL;

    /* Get the backend extensions string */
    if (pfnProxyQueryExtensionsString) {
        backend_extensions = pfnProxyQueryExtensionsString(
                                 display,
                                 DefaultScreen(display));
    } else if (pfnProxyEglQueryString) {
        backend_extensions = pfnProxyEglQueryString(
                                (EGLDisplay)display,
                                EGL_EXTENSIONS);
    }

    /* Get the GL extensions string */
    if (pfnProxyGetString)
        gl_extensions = (const char *)pfnProxyGetString(GL_EXTENSIONS);

    /* Check if present */
    if (glProxyCheckExtension(backend_extensions, extension) ||
        glProxyCheckExtension(gl_extensions, extension))
        return GL_TRUE;

    return GL_FALSE;
}

void *glProxyGetExtProcAddr(void *display, const char *extension, const char *name) {
    if (!glProxyHasExtension(display, extension))
        return NULL;

    if (pfnProxyEglGetProcAddress) {
        return (*pfnProxyEglGetProcAddress)(name);
    } else if (pfnProxyGLXGetProcAddress) {
        return (*pfnProxyGLXGetProcAddress)((const GLubyte *) name);
    } else if (pfnProxyGLXGetProcAddressARB) {
        return (*pfnProxyGLXGetProcAddressARB)((const GLubyte *) name);
    }

    return NULL;
}

static int glProxyCoreInit(void *handle)
{
	if (NULL == handle) {
        return -1;
    }

    pfnProxyActiveTexture = glProxyGetProcAddr(handle, "glActiveTexture");
    pfnProxyAttachShader = glProxyGetProcAddr(handle, "glAttachShader");
    pfnProxyBindAttribLocation = glProxyGetProcAddr(handle,
            "glBindAttribLocation");
    pfnProxyBindBuffer = glProxyGetProcAddr(handle, "glBindBuffer");
    pfnProxyBindFramebuffer = glProxyGetProcAddr(handle, "glBindFramebuffer");
    pfnProxyBindRenderbuffer = glProxyGetProcAddr(handle, "glBindRenderbuffer");
    pfnProxyBindTexture = glProxyGetProcAddr(handle, "glBindTexture");
    pfnProxyBlendColor = glProxyGetProcAddr(handle, "glBlendColor");
    pfnProxyBlendEquation = glProxyGetProcAddr(handle, "glBlendEquation");
    pfnProxyBlendEquationSeparate = glProxyGetProcAddr(handle,
            "glBlendEquationSeparate");
    pfnProxyBlendFunc = glProxyGetProcAddr(handle, "glBlendFunc");
    pfnProxyBlendFuncSeparate = glProxyGetProcAddr(handle,
            "glBlendFuncSeparate");
    pfnProxyBufferData = glProxyGetProcAddr(handle, "glBufferData");
    pfnProxyBufferSubData = glProxyGetProcAddr(handle, "glBufferSubData");
    pfnProxyCheckFramebufferStatus = glProxyGetProcAddr(handle,
            "glCheckFramebufferStatus");
    pfnProxyClear = glProxyGetProcAddr(handle, "glClear");
    pfnProxyClearColor = glProxyGetProcAddr(handle, "glClearColor");
    pfnProxyClearDepthf = glProxyGetProcAddr(handle, "glClearDepthf");
    pfnProxyClearStencil = glProxyGetProcAddr(handle, "glClearStencil");
    pfnProxyColorMask = glProxyGetProcAddr(handle, "glColorMask");
    pfnProxyCompileShader = glProxyGetProcAddr(handle, "glCompileShader");
    pfnProxyCompressedTexImage2D = glProxyGetProcAddr(handle,
            "glCompressedTexImage2D");
    pfnProxyCompressedTexSubImage2D = glProxyGetProcAddr(handle,
            "glCompressedTexSubImage2D");
    pfnProxyCopyTexImage2D = glProxyGetProcAddr(handle, "glCopyTexImage2D");
    pfnProxyCopyTexSubImage2D = glProxyGetProcAddr(handle,
            "glCopyTexSubImage2D");
    pfnProxyCreateProgram = glProxyGetProcAddr(handle, "glCreateProgram");
    pfnProxyCreateShader = glProxyGetProcAddr(handle, "glCreateShader");
    pfnProxyCullFace = glProxyGetProcAddr(handle, "glCullFace");
    pfnProxyDeleteBuffers = glProxyGetProcAddr(handle, "glDeleteBuffers");
    pfnProxyDeleteFramebuffers = glProxyGetProcAddr(handle,
            "glDeleteFramebuffers");
    pfnProxyDeleteProgram = glProxyGetProcAddr(handle, "glDeleteProgram");
    pfnProxyDeleteRenderbuffers = glProxyGetProcAddr(handle,
            "glDeleteRenderbuffers");
    pfnProxyDeleteShader = glProxyGetProcAddr(handle, "glDeleteShader");
    pfnProxyDeleteTextures = glProxyGetProcAddr(handle, "glDeleteTextures");
    pfnProxyDepthFunc = glProxyGetProcAddr(handle, "glDepthFunc");
    pfnProxyDepthMask = glProxyGetProcAddr(handle, "glDepthMask");
    pfnProxyDepthRangef = glProxyGetProcAddr(handle, "glDepthRangef");
    pfnProxyDetachShader = glProxyGetProcAddr(handle, "glDetachShader");
    pfnProxyDisable = glProxyGetProcAddr(handle, "glDisable");
    pfnProxyDisableVertexAttribArray = glProxyGetProcAddr(handle,
            "glDisableVertexAttribArray");
    pfnProxyDrawArrays = glProxyGetProcAddr(handle, "glDrawArrays");
    pfnProxyDrawElements = glProxyGetProcAddr(handle, "glDrawElements");
    pfnProxyEnable = glProxyGetProcAddr(handle, "glEnable");
    pfnProxyEnableVertexAttribArray = glProxyGetProcAddr(handle,
            "glEnableVertexAttribArray");
    pfnProxyFinish = glProxyGetProcAddr(handle, "glFinish");
    pfnProxyFlush = glProxyGetProcAddr(handle, "glFlush");
    pfnProxyFramebufferRenderbuffer = glProxyGetProcAddr(handle,
            "glFramebufferRenderbuffer");
    pfnProxyFramebufferTexture2D = glProxyGetProcAddr(handle,
            "glFramebufferTexture2D");
    pfnProxyFrontFace = glProxyGetProcAddr(handle, "glFrontFace");
    pfnProxyGenBuffers = glProxyGetProcAddr(handle, "glGenBuffers");
    pfnProxyGenerateMipmap = glProxyGetProcAddr(handle, "glGenerateMipmap");
    pfnProxyGenFramebuffers = glProxyGetProcAddr(handle, "glGenFramebuffers");
    pfnProxyGenRenderbuffers = glProxyGetProcAddr(handle, "glGenRenderbuffers");
    pfnProxyGenTextures = glProxyGetProcAddr(handle, "glGenTextures");
    pfnProxyGetActiveAttrib = glProxyGetProcAddr(handle, "glGetActiveAttrib");
    pfnProxyGetActiveUniform = glProxyGetProcAddr(handle, "glGetActiveUniform");
    pfnProxyGetAttachedShaders = glProxyGetProcAddr(handle,
            "glGetAttachedShaders");
    pfnProxyGetAttribLocation = glProxyGetProcAddr(handle,
            "glGetAttribLocation");
    pfnProxyGetBooleanv = glProxyGetProcAddr(handle, "glGetBooleanv");
    pfnProxyGetBufferParameteriv = glProxyGetProcAddr(handle,
            "glGetBufferParameteriv");
    pfnProxyGetError = glProxyGetProcAddr(handle, "glGetError");
    pfnProxyGetFloatv = glProxyGetProcAddr(handle, "glGetFloatv");
    pfnProxyGetFramebufferAttachmentParameteriv = glProxyGetProcAddr(handle,
            "glGetFramebufferAttachmentParameteriv");
    pfnProxyGetIntegerv = glProxyGetProcAddr(handle, "glGetIntegerv");
    pfnProxyGetProgramiv = glProxyGetProcAddr(handle, "glGetProgramiv");
    pfnProxyGetProgramInfoLog = glProxyGetProcAddr(handle,
            "glGetProgramInfoLog");
    pfnProxyGetRenderbufferParameteriv = glProxyGetProcAddr(handle,
            "glGetRenderbufferParameteriv");
    pfnProxyGetShaderiv = glProxyGetProcAddr(handle, "glGetShaderiv");
    pfnProxyGetShaderInfoLog = glProxyGetProcAddr(handle, "glGetShaderInfoLog");
    pfnProxyGetShaderPrecisionFormat = glProxyGetProcAddr(handle,
            "glGetShaderPrecisionFormat");
    pfnProxyGetShaderSource = glProxyGetProcAddr(handle, "glGetShaderSource");
    pfnProxyGetString = glProxyGetProcAddr(handle, "glGetString");
    pfnProxyGetTexParameterfv = glProxyGetProcAddr(handle,
            "glGetTexParameterfv");
    pfnProxyGetTexParameteriv = glProxyGetProcAddr(handle,
            "glGetTexParameteriv");
    pfnProxyGetUniformfv = glProxyGetProcAddr(handle, "glGetUniformfv");
    pfnProxyGetUniformiv = glProxyGetProcAddr(handle, "glGetUniformiv");
    pfnProxyGetUniformLocation = glProxyGetProcAddr(handle,
            "glGetUniformLocation");
    pfnProxyGetVertexAttribfv = glProxyGetProcAddr(handle,
            "glGetVertexAttribfv");
    pfnProxyGetVertexAttribiv = glProxyGetProcAddr(handle,
            "glGetVertexAttribiv");
    pfnProxyGetVertexAttribPointerv = glProxyGetProcAddr(handle,
            "glGetVertexAttribPointerv");
    pfnProxyHint = glProxyGetProcAddr(handle, "glHint");
    pfnProxyIsBuffer = glProxyGetProcAddr(handle, "glIsBuffer");
    pfnProxyIsEnabled = glProxyGetProcAddr(handle, "glIsEnabled");
    pfnProxyIsFramebuffer = glProxyGetProcAddr(handle, "glIsFramebuffer");
    pfnProxyIsProgram = glProxyGetProcAddr(handle, "glIsProgram");
    pfnProxyIsRenderbuffer = glProxyGetProcAddr(handle, "glIsRenderbuffer");
    pfnProxyIsShader = glProxyGetProcAddr(handle, "glIsShader");
    pfnProxyIsTexture = glProxyGetProcAddr(handle, "glIsTexture");
    pfnProxyLineWidth = glProxyGetProcAddr(handle, "glLineWidth");
    pfnProxyLinkProgram = glProxyGetProcAddr(handle, "glLinkProgram");
    pfnProxyPixelStorei = glProxyGetProcAddr(handle, "glPixelStorei");
    pfnProxyPolygonOffset = glProxyGetProcAddr(handle, "glPolygonOffset");
    pfnProxyReadPixels = glProxyGetProcAddr(handle, "glReadPixels");
    pfnProxyReleaseShaderCompiler = glProxyGetProcAddr(handle,
            "glReleaseShaderCompiler");
    pfnProxyRenderbufferStorage = glProxyGetProcAddr(handle,
            "glRenderbufferStorage");
    pfnProxySampleCoverage = glProxyGetProcAddr(handle, "glSampleCoverage");
    pfnProxyScissor = glProxyGetProcAddr(handle, "glScissor");
    pfnProxyShaderBinary = glProxyGetProcAddr(handle, "glShaderBinary");
    pfnProxyShaderSource = glProxyGetProcAddr(handle, "glShaderSource");
    pfnProxyStencilFunc = glProxyGetProcAddr(handle, "glStencilFunc");
    pfnProxyStencilFuncSeparate = glProxyGetProcAddr(handle,
            "glStencilFuncSeparate");
    pfnProxyStencilMask = glProxyGetProcAddr(handle, "glStencilMask");
    pfnProxyStencilMaskSeparate = glProxyGetProcAddr(handle,
            "glStencilMaskSeparate");
    pfnProxyStencilOp = glProxyGetProcAddr(handle, "glStencilOp");
    pfnProxyStencilOpSeparate = glProxyGetProcAddr(handle,
            "glStencilOpSeparate");
    pfnProxyTexImage2D = glProxyGetProcAddr(handle, "glTexImage2D");
    pfnProxyTexParameterf = glProxyGetProcAddr(handle, "glTexParameterf");
    pfnProxyTexParameterfv = glProxyGetProcAddr(handle, "glTexParameterfv");
    pfnProxyTexParameteri = glProxyGetProcAddr(handle, "glTexParameteri");
    pfnProxyTexParameteriv = glProxyGetProcAddr(handle, "glTexParameteriv");
    pfnProxyTexSubImage2D = glProxyGetProcAddr(handle, "glTexSubImage2D");
    pfnProxyUniform1f = glProxyGetProcAddr(handle, "glUniform1f");
    pfnProxyUniform1fv = glProxyGetProcAddr(handle, "glUniform1fv");
    pfnProxyUniform1i = glProxyGetProcAddr(handle, "glUniform1i");
    pfnProxyUniform1iv = glProxyGetProcAddr(handle, "glUniform1iv");
    pfnProxyUniform2f = glProxyGetProcAddr(handle, "glUniform2f");
    pfnProxyUniform2fv = glProxyGetProcAddr(handle, "glUniform2fv");
    pfnProxyUniform2i = glProxyGetProcAddr(handle, "glUniform2i");
    pfnProxyUniform2iv = glProxyGetProcAddr(handle, "glUniform2iv");
    pfnProxyUniform3f = glProxyGetProcAddr(handle, "glUniform3f");
    pfnProxyUniform3fv = glProxyGetProcAddr(handle, "glUniform3fv");
    pfnProxyUniform3i = glProxyGetProcAddr(handle, "glUniform3i");
    pfnProxyUniform3iv = glProxyGetProcAddr(handle, "glUniform3iv");
    pfnProxyUniform4f = glProxyGetProcAddr(handle, "glUniform4f");
    pfnProxyUniform4fv = glProxyGetProcAddr(handle, "glUniform4fv");
    pfnProxyUniform4i = glProxyGetProcAddr(handle, "glUniform4i");
    pfnProxyUniform4iv = glProxyGetProcAddr(handle, "glUniform4iv");
    pfnProxyUniformMatrix2fv = glProxyGetProcAddr(handle, "glUniformMatrix2fv");
    pfnProxyUniformMatrix3fv = glProxyGetProcAddr(handle, "glUniformMatrix3fv");
    pfnProxyUniformMatrix4fv = glProxyGetProcAddr(handle, "glUniformMatrix4fv");
    pfnProxyUseProgram = glProxyGetProcAddr(handle, "glUseProgram");
    pfnProxyValidateProgram = glProxyGetProcAddr(handle, "glValidateProgram");
    pfnProxyVertexAttrib1f = glProxyGetProcAddr(handle, "glVertexAttrib1f");
    pfnProxyVertexAttrib1fv = glProxyGetProcAddr(handle, "glVertexAttrib1fv");
    pfnProxyVertexAttrib2f = glProxyGetProcAddr(handle, "glVertexAttrib2f");
    pfnProxyVertexAttrib2fv = glProxyGetProcAddr(handle, "glVertexAttrib2fv");
    pfnProxyVertexAttrib3f = glProxyGetProcAddr(handle, "glVertexAttrib3f");
    pfnProxyVertexAttrib3fv = glProxyGetProcAddr(handle, "glVertexAttrib3fv");
    pfnProxyVertexAttrib4f = glProxyGetProcAddr(handle, "glVertexAttrib4f");
    pfnProxyVertexAttrib4fv = glProxyGetProcAddr(handle, "glVertexAttrib4fv");
    pfnProxyVertexAttribPointer = glProxyGetProcAddr(handle,
            "glVertexAttribPointer");
    pfnProxyViewport = glProxyGetProcAddr(handle, "glViewport");

    return 0;
}

extern int glxProxyInit(void *handle);
extern int eglProxyInit(void *handle);

glProxyContext *glProxyInit()
{
    glProxyContext *result = malloc(sizeof(glProxyContext));
    if (!result)
        return NULL;

    return result;
}

static const char *glProxyBackendDescription(GLuint backends)
{
    if (backends == GLPROXY_BACKEND_GLX_OPENGL)
        return "GLX OpenGL";
    else if (backends == GLPROXY_BACKEND_EGL_OPENGL)
        return "EGL OpenGL";
    else if (backends == GLPROXY_BACKEND_EGL_OPENGLES2)
        return "EGL OpenGL ES2.0";
    else if (backends == GLPROXY_BACKEND_NONE)
        return "None";
    else if (backends == GLPROXY_BACKEND_ANY)
        return "Any";

    return "Unknown";
}

GLuint glProxyCurrentBackends(glProxyContext *context)
{
    return context->current_backends;
}

static GLuint glProxyCheckBackends(GLuint backends)
{
    // Sanity checks
    // GLX only works with OpenGL
    if ((backends & GLPROXY_BACKEND_GLX) && !(backends & GLPROXY_BACKEND_OPENGL))
        return GLPROXY_BACKEND_NONE;

    return backends;
}

static GLuint glProxyBackendOrder[] = {
#ifdef __arm__  // EGL should be better on ARM
        GLPROXY_BACKEND_EGL_OPENGLES2,
        GLPROXY_BACKEND_GLX_OPENGL,
#else
        GLPROXY_BACKEND_GLX_OPENGL,
        GLPROXY_BACKEND_EGL_OPENGLES2,
#endif
        GLPROXY_BACKEND_EGL_OPENGL,
        GLPROXY_BACKEND_NONE
};

static GLuint glProxyGuessBestBackend(glProxyContext *context)
{
    GLuint i;
    // First, start with known backend infos
    if (context->num_detected_backends > 0) {
        // Choose the first one that's not software based
        for (i = 0; i < context->num_detected_backends; i++) {
            if (context->detected_backends[i].seems_hardware_renderer)
                return context->detected_backends[i].backend;
        }
    }

    // If backend detection hasn't been done, try libraries
    if (context->done_detect_libraries) {
        GLuint backend = GLPROXY_BACKEND_ANY;
        for (i = 0, backend = glProxyBackendOrder[0];
             backend != GLPROXY_BACKEND_NONE;
             i++, backend = glProxyBackendOrder[i]) {
            // If the libraries are present, select it
            if (context->detected_libraries & backend)
                break;
        }
        return backend;
    }

    // Nothing else to try, just do it in the default order
    return GLPROXY_BACKEND_ANY;
}

static GLboolean glProxyLoadGL(glProxyContext *context)
{
    context->gl_handle = glProxyOpenLibrary(LIBGL_NAME);
    if (!context->gl_handle)
        return GL_FALSE;

    context->current_backends |= GLPROXY_BACKEND_OPENGL;
    return GL_TRUE;
}

static GLboolean glProxyLoadGLES2(glProxyContext *context)
{
    context->gles2_handle = glProxyOpenLibrary(LIBGLES2_NAME);
    if (!context->gles2_handle)
        return GL_FALSE;

    context->current_backends |= GLPROXY_BACKEND_OPENGLES2;
    return GL_TRUE;
}

static GLboolean glProxyLoadEGL(glProxyContext *context)
{
    pfnProxyEglGetProcAddress = NULL;

    context->egl_handle = glProxyOpenLibrary(LIBEGL_NAME);
    if (!context->egl_handle)
        return GL_FALSE;
    pfnProxyEglGetProcAddress = dlsym(context->egl_handle, "eglGetProcAddress");
    if (pfnProxyEglGetProcAddress == NULL) {
        dlclose(context->egl_handle);
        return GL_FALSE;
    }

    context->current_backends |= GLPROXY_BACKEND_EGL;
    return GL_TRUE;
}

static GLboolean glProxyLoadGLX(glProxyContext *context)
{
    pfnProxyGLXGetProcAddress = NULL;
    pfnProxyGLXGetProcAddressARB = NULL;

    if (!glProxyLoadGL(context))
        return GL_FALSE;

    pfnProxyGLXGetProcAddress = dlsym(context->gl_handle, "glXGetProcAddress");
    if (pfnProxyGLXGetProcAddress == NULL) {
        pfnProxyGLXGetProcAddressARB = dlsym(context->gl_handle,
                "glXGetProcAddressARB");
        if (pfnProxyGLXGetProcAddressARB == NULL) {
            dlclose(context->gl_handle);
            context->current_backends = GLPROXY_BACKEND_NONE;
            return GL_FALSE;
        }
    }

    context->current_backends |= GLPROXY_BACKEND_GLX;
    return GL_TRUE;
}

GLboolean glProxyLoad(glProxyContext *context, GLuint backends)
{
    if (!context || context->current_backends != GLPROXY_BACKEND_NONE)
        // Not present or already loaded context
        return GL_FALSE;

    printf("%s: Backend requested: %s\n", __FUNCTION__, glProxyBackendDescription(backends));

    // Check sanity of request
    GLuint backends_toload = glProxyCheckBackends(backends);
    if (backends_toload == GLPROXY_BACKEND_NONE)
        return GL_FALSE;

    // Select one if none passed
    if (backends_toload == GLPROXY_BACKEND_ANY)
        backends_toload = glProxyGuessBestBackend(context);

    // Now try to load the libraries (in order)
    GLuint backend_try, i;
    for (i = 0, backend_try = glProxyBackendOrder[0];
         backend_try != GLPROXY_BACKEND_NONE;
         i++, backend_try = glProxyBackendOrder[i]) {
        // If one was selected, and it's not this one, pass
        if ((backends_toload != backend_try) &&
            (backends_toload != GLPROXY_BACKEND_ANY))
            continue;

        if (backend_try & GLPROXY_BACKEND_EGL) {
            if (!glProxyLoadEGL(context))
                continue;

            if (backend_try & GLPROXY_BACKEND_OPENGLES2) {
                if (!glProxyLoadGLES2(context)) {
                    glProxyCloseLibrary(context->egl_handle);
                    continue;
                }
                break;
            }

            if (backend_try & GLPROXY_BACKEND_OPENGL) {
                if (!glProxyLoadGL(context)) {
                    glProxyCloseLibrary(context->egl_handle);
                    continue;
                }
                break;
            }
        }

        if (backend_try & GLPROXY_BACKEND_GLX) {
            if (!glProxyLoadGLX(context))
                continue;
            break;
        }

    }

    if (!context->current_backends)
        return GL_FALSE;

    // Initialize all the symbols
    if (context->current_backends & GLPROXY_BACKEND_EGL)
        eglProxyInit(context->egl_handle);
    if (context->current_backends & GLPROXY_BACKEND_GLX)
        glxProxyInit(context->gl_handle);
    if (backends_toload & GLPROXY_BACKEND_OPENGL)
        glProxyCoreInit(context->gl_handle);
    if (backends_toload & GLPROXY_BACKEND_OPENGLES2)
        glProxyCoreInit(context->gles2_handle);

    printf("%s: Backend loaded: %s\n", __FUNCTION__, glProxyBackendDescription(context->current_backends));

    return GL_TRUE;
}

void glProxyClose(glProxyContext *context)
{
    if (!context)
        return;

    if (context->gl_handle) {
        glProxyCloseLibrary(context->gl_handle);
        context->gl_handle = NULL;
    }

    if (context->gles2_handle) {
        glProxyCloseLibrary(context->gles2_handle);
        context->gles2_handle = NULL;
    }

    if (context->egl_handle) {
        glProxyCloseLibrary(context->egl_handle);
        context->egl_handle = NULL;
    }

    context->current_backends = GLPROXY_BACKEND_NONE;

    GLuint i;
    for (i = 0; i < context->num_detected_backends; i++) {
        if (context->detected_backends[i].version)
            free((void *)context->detected_backends[i].version);
        if (context->detected_backends[i].vendor)
            free((void *)context->detected_backends[i].vendor);
        if (context->detected_backends[i].renderer)
            free((void *)context->detected_backends[i].renderer);
        if (context->detected_backends[i].backend_extensions)
            free((void *)context->detected_backends[i].backend_extensions);
        if (context->detected_backends[i].gl_extensions)
            free((void *)context->detected_backends[i].gl_extensions);
    }

    free(context);
}

GLuint glProxyDetectLibraries(glProxyContext *context)
{
    void *main_handle = NULL;
    Dl_serinfo infosize, *info = NULL;
    Dl_serpath *path;
    unsigned int count;
    char filename[4096];

    if (context->done_detect_libraries) {
        context->done_detect_libraries = GL_FALSE;
        context->detected_libraries = 0;
    }

    main_handle = dlopen(0, RTLD_NOW|RTLD_GLOBAL);

    dlinfo(main_handle, RTLD_DI_SERINFOSIZE, (void *)&infosize);

    info = malloc(infosize.dls_size);
    info->dls_size = infosize.dls_size;
    info->dls_cnt = infosize.dls_cnt;
    dlinfo(main_handle, RTLD_DI_SERINFO, (void *)info);
    path = &info->dls_serpath[0];

    for (count = 1; count <= info->dls_cnt; count++, path++) {
        // Test for libGL.so.1
        snprintf(filename, 4096, "%s/%s", path->dls_name, LIBGL_NAME);
        if (euidaccess(filename, R_OK) == 0) {
            // Consider libGL to implement both GLX and desktop OpenGL
            context->detected_libraries |= GLPROXY_BACKEND_GLX;
            context->detected_libraries |= GLPROXY_BACKEND_OPENGL;
        } else {
            // Also test for libGL.so, as it's a symlink in some systems
            snprintf(filename, 4096, "%s/%s", path->dls_name, "libGL.so");
            if (euidaccess(filename, R_OK) == 0) {
                // Consider libGL to implement both GLX and desktop OpenGL
                context->detected_libraries |= GLPROXY_BACKEND_GLX;
                context->detected_libraries |= GLPROXY_BACKEND_OPENGL;
            }
        }

        // Test for libEGL.so
        snprintf(filename, 4096, "%s/%s", path->dls_name, LIBEGL_NAME);
        if (euidaccess(filename, R_OK) == 0) {
            context->detected_libraries |= GLPROXY_BACKEND_EGL;
        }

        // Test for libGLESv2.so
        snprintf(filename, 4096, "%s/%s", path->dls_name, LIBGLES2_NAME);
        if (euidaccess(filename, R_OK) == 0) {
            context->detected_libraries |= GLPROXY_BACKEND_OPENGLES2;
        }
    }

    context->done_detect_libraries = GL_TRUE;
    return context->detected_libraries;
}

static void glProxyFillBackendInfo(glProxyBackendInfo *info,
                                   GLuint backend,
                                   const char *vendor,
                                   const char *renderer,
                                   const char *version,
                                   const char *backend_extensions,
                                   const char *gl_extensions)
{
    info->backend = backend;
    info->description = malloc(strlen(glProxyBackendDescription(backend)) + 1);
    strcpy((char *)info->description, glProxyBackendDescription(backend));
    info->vendor = malloc(strlen(vendor)+1);
    strcpy((char *)info->vendor, vendor);
    info->renderer = malloc(strlen(renderer)+1);
    strcpy((char *)info->renderer, renderer);
    info->version = malloc(strlen(version)+1);
    strcpy((char *)info->version, version);
    info->backend_extensions = malloc(strlen(backend_extensions)+1);
    strcpy((char *)info->backend_extensions, backend_extensions);
    info->gl_extensions = malloc(strlen(gl_extensions)+1);
    strcpy((char *)info->gl_extensions, gl_extensions);

    // Test to see if the renderer is hardware based.
    if (!strcasestr(info->renderer, "soft")) {
        info->seems_hardware_renderer = GL_TRUE;
    }
}

static void glxProxyDetectBackends(glProxyContext *context)
{
    Display *x_display;
    int scrnum;
    Window root;
    GLXContext glx_ctx = 0;
    int attribs[] = {
            GLX_RGBA,
            None };
    XVisualInfo *visinfo = NULL;

    if (!context->gl_handle)
        if (!glProxyLoadGLX(context))
            return;

    pfnProxyChooseVisual = glProxyGetProcAddr(context->gl_handle, "glXChooseVisual");
    pfnProxyCreateContext = glProxyGetProcAddr(context->gl_handle, "glXCreateContext");
    pfnProxyDestroyContext = glProxyGetProcAddr(context->gl_handle, "glXDestroyContext");
    pfnProxyMakeCurrent = glProxyGetProcAddr(context->gl_handle, "glXMakeCurrent");
    pfnProxyGetString = glProxyGetProcAddr(context->gl_handle, "glGetString");
    pfnProxyQueryExtensionsString = glProxyGetProcAddr(context->gl_handle,
            "glXQueryExtensionsString");

    x_display = XOpenDisplay(NULL);
    if (x_display == NULL)
        goto cleanup;

    scrnum = DefaultScreen(x_display);
    root = RootWindow(x_display, scrnum);
    visinfo = pfnProxyChooseVisual(x_display, scrnum, attribs);
    if (!visinfo)
        goto cleanup;
    glx_ctx = pfnProxyCreateContext(x_display, visinfo, NULL, True);
    if (!glx_ctx)
        goto cleanup;

    pfnProxyMakeCurrent(x_display, root, glx_ctx);
    glProxyFillBackendInfo(&context->detected_backends[context->num_detected_backends],
                           GLPROXY_BACKEND_GLX_OPENGL,
                           (const char *) pfnProxyGetString(GL_VENDOR),
                           (const char *) pfnProxyGetString(GL_RENDERER),
                           (const char *) pfnProxyGetString(GL_VERSION),
                           (const char *) pfnProxyQueryExtensionsString(x_display, scrnum),
                           (const char *) pfnProxyGetString(GL_EXTENSIONS));
    context->num_detected_backends++;

    pfnProxyMakeCurrent(x_display, None, NULL);
cleanup:
    if (glx_ctx)
        pfnProxyDestroyContext(x_display, glx_ctx);
    if (x_display)
        XCloseDisplay(x_display);
    pfnProxyGetString = NULL;
    pfnProxyChooseVisual = NULL;
    pfnProxyCreateContext = NULL;
    pfnProxyMakeCurrent = NULL;
    pfnProxyDestroyContext = NULL;
    pfnProxyQueryExtensionsString = NULL;
    if (context->gl_handle) {
        glProxyCloseLibrary(context->gl_handle);
        context->gl_handle = NULL;
        context->current_backends &= ~GLPROXY_BACKEND_GLX_OPENGL;
    }
    pfnProxyGLXGetProcAddress = NULL;
    pfnProxyGLXGetProcAddressARB = NULL;
}

static void eglProxyDetectBackendGLES2(glProxyContext *context,
                                       EGLDisplay egl_display)
{
    EGLConfig egl_config;
    EGLint num_configs = 1;
    EGLContext egl_ctx = EGL_NO_CONTEXT;
    EGLSurface egl_surface = EGL_NO_SURFACE;

    EGLint config_attribs[] = {
            EGL_SURFACE_TYPE, EGL_PBUFFER_BIT,
            EGL_RENDERABLE_TYPE, EGL_OPENGL_ES2_BIT,
            EGL_NONE
    };
    EGLint ctx_attribs[] = {
            EGL_CONTEXT_CLIENT_VERSION, 2,
            EGL_NONE
    };
    static const EGLint surface_attribs[] = {
            EGL_WIDTH, 1,
            EGL_HEIGHT, 1,
            EGL_NONE
    };

    if (!context->gles2_handle)
        if (!glProxyLoadGLES2(context))
            return;

    pfnProxyGetString = glProxyGetProcAddr(context->gles2_handle, "glGetString");
    if (!pfnProxyGetString)
        goto cleanup;

    if (!pfnProxyEglChooseConfig(egl_display, config_attribs, &egl_config,
                                 num_configs, &num_configs) ||
        num_configs < 1)
        goto cleanup;
    if(pfnProxyEglBindAPI &&
       (pfnProxyEglBindAPI(EGL_OPENGL_ES_API) != EGL_TRUE))
        goto cleanup;
    egl_ctx = pfnProxyEglCreateContext(egl_display, egl_config, EGL_NO_CONTEXT, ctx_attribs);
    if (egl_ctx == EGL_NO_CONTEXT)
        goto cleanup;
    egl_surface = pfnProxyEglCreatePbufferSurface(egl_display, egl_config, surface_attribs);
    if (egl_surface == EGL_NO_SURFACE)
        goto cleanup;
    if (!pfnProxyEglMakeCurrent(egl_display, egl_surface, egl_surface, egl_ctx))
        goto cleanup;

    glProxyFillBackendInfo(&context->detected_backends[context->num_detected_backends],
                           GLPROXY_BACKEND_EGL_OPENGLES2,
                           (const char *) pfnProxyGetString(GL_VENDOR),
                           (const char *) pfnProxyGetString(GL_RENDERER),
                           (const char *) pfnProxyGetString(GL_VERSION),
                           (const char *) pfnProxyEglQueryString(egl_display, EGL_EXTENSIONS),
                           (const char *) pfnProxyGetString(GL_EXTENSIONS));
    context->num_detected_backends++;

    pfnProxyEglMakeCurrent(egl_display, EGL_NO_SURFACE, EGL_NO_SURFACE, EGL_NO_CONTEXT);

cleanup:
    if (egl_surface != EGL_NO_SURFACE)
        pfnProxyEglDestroySurface(egl_display, egl_surface);
    if (egl_ctx != EGL_NO_CONTEXT)
        pfnProxyEglDestroyContext(egl_display, egl_ctx);
    pfnProxyGetString = NULL;
    if (context->gles2_handle) {
        glProxyCloseLibrary(context->gles2_handle);
        context->gles2_handle = NULL;
        context->current_backends &= ~GLPROXY_BACKEND_OPENGLES2;
    }
}

static void eglProxyDetectBackendGL(glProxyContext *context,
                                    EGLDisplay egl_display)
{
    EGLConfig egl_config;
    EGLint num_configs = 1;
    EGLContext egl_ctx = EGL_NO_CONTEXT;
    EGLSurface egl_surface = EGL_NO_SURFACE;

    EGLint config_attribs[] = {
            EGL_SURFACE_TYPE, EGL_PBUFFER_BIT,
            EGL_RENDERABLE_TYPE, EGL_OPENGL_BIT,
            EGL_NONE
    };
    EGLint ctx_attribs[] = {
            EGL_NONE
    };
    static const EGLint surface_attribs[] = {
            EGL_WIDTH, 1,
            EGL_HEIGHT, 1,
            EGL_NONE
    };

    if (!context->gl_handle)
        if (!glProxyLoadGL(context))
            return;

    pfnProxyGetString = glProxyGetProcAddr(context->gl_handle, "glGetString");
    if (!pfnProxyGetString)
        goto cleanup;

    if (!pfnProxyEglChooseConfig(egl_display, config_attribs, &egl_config,
                                 num_configs, &num_configs) ||
        num_configs < 1)
        goto cleanup;
    if(pfnProxyEglBindAPI &&
       (pfnProxyEglBindAPI(EGL_OPENGL_API) != EGL_TRUE))
        goto cleanup;
    egl_ctx = pfnProxyEglCreateContext(egl_display, egl_config, EGL_NO_CONTEXT, ctx_attribs);
    if (egl_ctx == EGL_NO_CONTEXT)
        goto cleanup;
    egl_surface = pfnProxyEglCreatePbufferSurface(egl_display, egl_config, surface_attribs);
    if (egl_surface == EGL_NO_SURFACE)
        goto cleanup;
    if (!pfnProxyEglMakeCurrent(egl_display, egl_surface, egl_surface, egl_ctx))
        goto cleanup;

    glProxyFillBackendInfo(&context->detected_backends[context->num_detected_backends],
                           GLPROXY_BACKEND_EGL_OPENGL,
                           (const char *) pfnProxyGetString(GL_VENDOR),
                           (const char *) pfnProxyGetString(GL_RENDERER),
                           (const char *) pfnProxyGetString(GL_VERSION),
                           (const char *) pfnProxyEglQueryString(egl_display, EGL_EXTENSIONS),
                           (const char *) pfnProxyGetString(GL_EXTENSIONS));
    context->num_detected_backends++;

    pfnProxyEglMakeCurrent(egl_display, EGL_NO_SURFACE, EGL_NO_SURFACE, EGL_NO_CONTEXT);

cleanup:
    if (egl_surface != EGL_NO_SURFACE)
        pfnProxyEglDestroySurface(egl_display, egl_surface);
    if (egl_ctx != EGL_NO_CONTEXT)
        pfnProxyEglDestroyContext(egl_display, egl_ctx);
    pfnProxyGetString = NULL;
    if (context->gl_handle) {
        glProxyCloseLibrary(context->gl_handle);
        context->gl_handle = NULL;
        context->current_backends &= ~GLPROXY_BACKEND_OPENGL;
    }
}

static void eglProxyDetectBackends(glProxyContext *context)
{
    Display *x_display = NULL;
    EGLDisplay egl_display = EGL_NO_DISPLAY;

    if (!context->egl_handle)
        if (!glProxyLoadEGL(context))
            return;

    pfnProxyEglGetDisplay = glProxyGetProcAddr(context->egl_handle, "eglGetDisplay");
    pfnProxyEglInitialize = glProxyGetProcAddr(context->egl_handle, "eglInitialize");
    pfnProxyEglChooseConfig = glProxyGetProcAddr(context->egl_handle, "eglChooseConfig");
    pfnProxyEglCreateContext = glProxyGetProcAddr(context->egl_handle, "eglCreateContext");
    pfnProxyEglCreatePbufferSurface = glProxyGetProcAddr(context->egl_handle, "eglCreatePbufferSurface");
    pfnProxyEglMakeCurrent = glProxyGetProcAddr(context->egl_handle, "eglMakeCurrent");
    pfnProxyEglDestroyContext = glProxyGetProcAddr(context->egl_handle, "eglDestroyContext");
    pfnProxyEglDestroySurface = glProxyGetProcAddr(context->egl_handle, "eglDestroySurface");
    pfnProxyEglTerminate = glProxyGetProcAddr(context->egl_handle, "eglTerminate");
    pfnProxyEglQueryString = glProxyGetProcAddr(context->egl_handle, "eglQueryString");
    pfnProxyEglBindAPI = glProxyGetProcAddr(context->egl_handle, "eglBindAPI");

    x_display = XOpenDisplay(NULL);
    if (x_display == NULL)
        goto cleanup;
    egl_display = pfnProxyEglGetDisplay(x_display);
    if ((egl_display == EGL_NO_DISPLAY) ||
        !pfnProxyEglInitialize(egl_display, NULL, NULL))
        goto cleanup;

    eglProxyDetectBackendGLES2(context, egl_display);
    eglProxyDetectBackendGL(context, egl_display);

cleanup:
    if (egl_display != EGL_NO_DISPLAY)
        pfnProxyEglTerminate(egl_display);
    if (x_display)
        XCloseDisplay(x_display);
    pfnProxyEglGetDisplay = NULL;
    pfnProxyEglInitialize = NULL;
    pfnProxyEglChooseConfig = NULL;
    pfnProxyEglCreateContext = NULL;
    pfnProxyEglCreatePbufferSurface = NULL;
    pfnProxyEglMakeCurrent = NULL;
    pfnProxyEglDestroyContext = NULL;
    pfnProxyEglDestroySurface = NULL;
    pfnProxyEglTerminate = NULL;
    pfnProxyEglQueryString = NULL;
    pfnProxyEglBindAPI = NULL;
    if (context->egl_handle) {
        glProxyCloseLibrary(context->egl_handle);
        context->egl_handle = NULL;
        context->current_backends = ~GLPROXY_BACKEND_EGL;
    }
    pfnProxyEglGetProcAddress = NULL;
}

glProxyBackendInfo *
glProxyDetectBackends(glProxyContext *context, GLuint *num_backends)
{
    if (context->current_backends != 0)
        return NULL;

    glxProxyDetectBackends(context);
    eglProxyDetectBackends(context);
    if (num_backends)
        *num_backends = context->num_detected_backends;
    return context->detected_backends;
}
