/*
 * glproxy list backends utility.
 *
 * Copyright (C) 2012 Linaro Ltd.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Authors:
 *  Marc Ordinas i Llopis, Collabora Ltd <marc.ordinasillopis@linaro.org>
 */


#include <stdlib.h>
#include <stdio.h>
#include <GL/gl_proxy.h>

int main(int argc, char *argv[])
{
    glProxyContext *context;
    GLuint backends;
    glProxyBackendInfo *infos;
    GLuint num_infos, i;

    context = glProxyInit();
    backends = glProxyDetectLibraries(context);
    printf("Libraries detected: ");
    if (backends & GLPROXY_BACKEND_EGL)
        printf("EGL ");
    if (backends & GLPROXY_BACKEND_GLX)
        printf("GLX ");
    if (backends & GLPROXY_BACKEND_OPENGL)
        printf("OpenGL ");
    if (backends & GLPROXY_BACKEND_OPENGLES2)
        printf("OpenGLES2");
    printf("\n");

    infos = glProxyDetectBackends(context, &num_infos);
    printf("Backends found: %d\n", num_infos);
    for (i = 0; i < num_infos; i++) {
        printf("Backend %s (0x%04x):\n", infos[i].description, infos[i].backend);
        printf("\tVendor: %s\n", infos[i].vendor);
        printf("\tVersion: %s\n", infos[i].version);
        printf("\tRenderer: %s (seems %s based)\n", infos[i].renderer,
               infos[i].seems_hardware_renderer? "hardware" : "software");
        printf("\tBackend extensions: %s\n", infos[i].backend_extensions);
        printf("\tGL extensions: %s\n", infos[i].gl_extensions);
        printf("\n");
    }
    return 0;
}
